// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:meta/meta.dart';
import 'package:wyatt_form_bloc/src/enums/enums.dart';
import 'package:wyatt_form_bloc/src/form/form.dart';
import 'package:wyatt_form_bloc/src/utils/list_equals.dart';

@immutable
class FormData {
  const FormData(this._entries);

  FormData.empty() : this(<FormEntry>[]);

  final List<FormEntry> _entries;

  List<FormEntry> get entries => _entries;

  List<FormInput<T, ValidationError>> inputs<T>() {
    return _entries
        .map((FormEntry entry) => entry.input as FormInput<T, ValidationError>)
        .toList();
  }

  static FormStatus validate(List<FormInput> inputs) {
    return inputs.every((FormInput element) => element.pure)
        ? FormStatus.pure
        : inputs.any((FormInput input) => input.valid == false)
            ? FormStatus.invalid
            : FormStatus.valid;
  }

  FormStatus selfValidate() {
    return validate(inputs<dynamic>());
  }

  FormInput<T, ValidationError> input<T>(String key) {
    if (contains(key)) {
      return _entries.firstWhere((FormEntry entry) => entry.key == key).input
          as FormInput<T, ValidationError>;
    } else {
      throw Exception('FormInput with key `$key` does not exist in form');
    }
  }

  bool contains(String key) {
    return _entries.any((FormEntry entry) => entry.key == key);
  }

  FormData intersection(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry entry in _entries) {
      if (other.contains(entry.key)) {
        entries.add(entry);
      }
    }

    return FormData(entries);
  }

  FormData difference(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry otherEntry in other._entries) {
      if (!contains(otherEntry.key)) {
        entries.add(otherEntry);
      }
    }

    for (final FormEntry entry in _entries) {
      if (!other.contains(entry.key)) {
        entries.add(entry);
      }
    }

    return FormData(entries);
  }

  FormData union(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry entry in _entries) {
      entries.add(entry);
    }

    for (final FormEntry otherEntry in other._entries) {
      if (!contains(otherEntry.key)) {
        entries.add(otherEntry);
      }
    }

    return FormData(entries);
  }

  void update(String key, FormInput input) {
    if (contains(key)) {
      final index = _entries.indexOf(
        _entries.firstWhere((FormEntry entry) => entry.key == key),
      );
      _entries[index] = _entries[index].copyWith(input: input);
    }
  }

  FormData clone() {
    return FormData(
      _entries.map((FormEntry entry) => entry.clone()).toList(),
    );
  }

  Map<String, dynamic> toMap() {
    final map = <String, dynamic>{};
    for (final entry in _entries) {
      if (entry.export) {
        map[entry.name] = entry.input.value;
      }
    }
    return map;
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is FormData && listEquals(other._entries, _entries);
  }

  @override
  int get hashCode => _entries.hashCode;

  @override
  String toString() => 'FormData(entries: $_entries)';
}
