// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';
import 'dart:typed_data';

import 'package:http/http.dart';
import 'package:wyatt_http_client/src/mixins/body_transformer.dart';
import 'package:wyatt_http_client/src/mixins/headers_transformer.dart';
import 'package:wyatt_http_client/src/rest_client.dart';
import 'package:wyatt_http_client/src/utils/protocols.dart';
import 'package:wyatt_http_client/src/utils/utils.dart';

class AwesomeClient extends BaseClient {
  final Client _inner;

  AwesomeClient({
    Client? inner,
  }) : _inner = inner ?? Client();

  @override
  Future<StreamedResponse> send(BaseRequest request) {
    return _inner.send(request);
  }
}

class AwesomeRestClient extends AwesomeClient
    with HeadersTransformer, BodyTransformer {
  final Protocols protocol;
  final String? authority;

  AwesomeRestClient({
    this.protocol = Protocols.https,
    this.authority = '',
    super.inner,
  });

  @override
  Object? bodyMutator(Object? body) {
    print('bodyMutator: Json encoding');
    return jsonEncode(body);
  }

  @override
  Map<String, String>? headersMutator(
    Object? body,
    Map<String, String>? headers,
  ) {
    print('headerMutator: Json encoding');
    final mutation = {
      'content-type': 'application/json; charset=utf-8',
    };
    if (headers != null) {
      headers.addAll(mutation);
      return headers;
    } else {
      return mutation;
    }
  }

  @override
  BaseRequest requestMutator(BaseRequest request) {
    print('requestMutator: scheme + authority');
    final Uri uri = Uri.parse('${protocol.scheme}$authority${request.url}');
    return Utils.copyRequestWith(request, url: uri);
  }
}

class AwesomeUrlClient extends AuthenticatedClient {
  AwesomeUrlClient(super.inner);
}

class AwesomeOauth2Client extends AuthenticatedClient with HeadersTransformer {
  AwesomeOauth2Client(super.inner);

  @override
  Map<String, String>? headersMutator(
    Object? body,
    Map<String, String>? headers,
  ) {
    print('headersMutator: Token manager');
    final mutation = {
      'authorization': 'Bearer TOKEN',
    };
    if (headers != null) {
      headers.addAll(mutation);
      return headers;
    } else {
      return mutation;
    }
  }
}

abstract class AuthenticatedClient implements Client {
  final Client _inner;

  Client get inner => _inner;

  AuthenticatedClient(BaseClient? inner) : _inner = inner ?? RestClient();

  @override
  void close() => _inner.close();

  @override
  Future<Response> head(Uri url, {Map<String, String>? headers}) =>
      _inner.head(url, headers: headers);

  @override
  Future<Response> get(Uri url, {Map<String, String>? headers}) =>
      _inner.get(url, headers: headers);

  @override
  Future<Response> post(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.post(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> put(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.put(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> patch(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.patch(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> delete(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.delete(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<String> read(Uri url, {Map<String, String>? headers}) =>
      _inner.read(url, headers: headers);

  @override
  Future<Uint8List> readBytes(Uri url, {Map<String, String>? headers}) =>
      _inner.readBytes(url, headers: headers);

  @override
  Future<StreamedResponse> send(BaseRequest request) => _inner.send(request);
}

abstract class AuthenticationClient extends BaseClient {
  final BaseClient _inner;

  BaseClient get inner => _inner;

  AuthenticationClient(BaseClient? inner) : _inner = inner ?? RestClient();

  @override
  Future<Response> head(Uri url, {Map<String, String>? headers}) =>
      _inner.head(url, headers: headers);

  @override
  Future<Response> get(Uri url, {Map<String, String>? headers}) =>
      _inner.get(url, headers: headers);

  @override
  Future<Response> post(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.post(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> put(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.put(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> patch(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.patch(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<Response> delete(
    Uri url, {
    Map<String, String>? headers,
    Object? body,
    Encoding? encoding,
  }) =>
      _inner.delete(url, headers: headers, body: body, encoding: encoding);

  @override
  Future<StreamedResponse> send(BaseRequest request) {
    return _inner.send(request);
  }

  @override
  void close() {
    _inner.close();
    return super.close();
  }
}
