// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:http/http.dart';
import 'package:wyatt_http_client/src/middleware_client.dart';
import 'package:wyatt_http_client/src/models/middleware_request.dart';
import 'package:wyatt_http_client/src/models/middleware_response.dart';
import 'package:wyatt_http_client/src/pipeline.dart';
import 'package:wyatt_http_client/src/utils/authentication_methods.dart';
import 'package:wyatt_http_client/src/utils/header_keys.dart';
import 'package:wyatt_http_client/src/utils/http_status.dart';

typedef TokenParser = String Function(Map<String, dynamic>);

class RefreshTokenMiddleware extends Middleware {
  final String authorizationEndpoint;
  final String tokenEndpoint;

  String? accessToken;
  final TokenParser accessTokenParser;
  String? refreshToken;
  final TokenParser refreshTokenParser;

  final String authenticationHeader;
  final String authenticationMethod;
  final HttpStatus unauthorized;
  final int maxRetries;

  RefreshTokenMiddleware({
    required this.authorizationEndpoint,
    required this.tokenEndpoint,
    required this.accessTokenParser,
    required this.refreshTokenParser,
    this.authenticationHeader = HeaderKeys.authorization,
    this.authenticationMethod = AuthenticationMethods.bearer,
    this.unauthorized = HttpStatus.unauthorized,
    this.maxRetries = 3,
  });

  @override
  String getName() => 'RefreshToken';

  @override
  Future<MiddlewareRequest> onRequest(MiddlewareRequest request) async {
    print(
      '${getName()}::OnRequest: accessToken: $accessToken',
    );
    if (request.context.originalRequest?.unfreezedRequest.url ==
        Uri.parse(authorizationEndpoint)) {
      return super.onRequest(request);
    }
    if (accessToken != null) {
      // Modify header with accessToken
      var newReq = request.unfreezedRequest;
      final mutation = {
        authenticationHeader: '$authenticationMethod $accessToken',
      };
      Map<String, String>? headers = newReq.headers;
      if (headers != null) {
        headers.addAll(mutation);
      } else {
        headers = mutation;
      }
      newReq = newReq.copyWith(headers: headers);
      request.updateUnfreezedRequest(newReq);
      return super.onRequest(request);
    }
    if (refreshToken != null) {
      // Refresh accessToken with refreshToken before perform request
      final subPipeline = request.context.pipeline.fromUntil(this);
      final httpClient = MiddlewareClient(
        pipeline: subPipeline,
        inner: client,
      );
      final Map<String, String> headers = {
        authenticationHeader: '$authenticationHeader $refreshToken',
      };
      final response =
          await httpClient.get(Uri.parse(tokenEndpoint), headers: headers);
      final status = HttpStatus.from(response.statusCode);
      if (status.isSuccess()) {
        final body = jsonDecode(response.body) as Map<String, dynamic>;
        accessToken = accessTokenParser(body);

        // Then modify current request with accessToken
        var newReq = request.unfreezedRequest;
        final mutation = {
          authenticationHeader: '$authenticationMethod $accessToken',
        };
        Map<String, String>? headers = newReq.headers;
        if (headers != null) {
          headers.addAll(mutation);
        } else {
          headers = mutation;
        }
        newReq = newReq.copyWith(headers: headers);
        request.updateUnfreezedRequest(newReq);
      } else {
        // Retry
        int retries = 0;
        while (retries < maxRetries) {
          final Map<String, String> headers = {
            authenticationHeader: '$authenticationHeader $refreshToken',
          };
          final response =
              await httpClient.get(Uri.parse(tokenEndpoint), headers: headers);
          final status = HttpStatus.from(response.statusCode);
          if (status.isSuccess()) {
            final body = jsonDecode(response.body) as Map<String, dynamic>;
            accessToken = accessTokenParser(body);

            // Then modify current request with accessToken
            var newReq = request.unfreezedRequest;
            final mutation = {
              authenticationHeader: '$authenticationMethod $accessToken',
            };
            Map<String, String>? headers = newReq.headers;
            if (headers != null) {
              headers.addAll(mutation);
            } else {
              headers = mutation;
            }
            newReq = newReq.copyWith(headers: headers);
            request.updateUnfreezedRequest(newReq);
            break;
          }
          retries++;
        }
      }
      return super.onRequest(request);
    }
    // Pass
    return super.onRequest(request);
  }

  @override
  Future<MiddlewareResponse> onResponse(MiddlewareResponse response) async {
    final res = await super.onResponse(response);
    final status = HttpStatus.from(res.httpResponse.statusCode);
    if (res.context.originalRequest?.unfreezedRequest.url ==
        Uri.parse(authorizationEndpoint)) {
      if (status.isSuccess()) {
        final body = jsonDecode((res.httpResponse as Response).body) as Map<String, dynamic>;
        final accessToken = accessTokenParser(body);
        final refreshToken = refreshTokenParser(body);

        if (accessToken.isNotEmpty) {
          this.accessToken = accessToken;
        }
        if (refreshToken.isNotEmpty) {
          this.refreshToken = refreshToken;
        }
      }
      return res;
    }
    if (status == unauthorized) {
      print(
        '${getName()}::OnResponse: $unauthorized',
      );
      // Refresh token then retry
      final subPipeline = res.context.pipeline.fromUntil(this);
      final httpClient = MiddlewareClient(
        pipeline: subPipeline,
        inner: client,
      );
      final Map<String, String> headers = {
        authenticationHeader: '$authenticationHeader $refreshToken',
      };
      final response =
          await httpClient.get(Uri.parse(tokenEndpoint), headers: headers);
      final refreshstatus = HttpStatus.from(response.statusCode);
      if (refreshstatus.isSuccess()) {
        print(
          '${getName()}::OnResponse: refresh successfuly',
        );
        final body = jsonDecode(response.body) as Map<String, dynamic>;
        accessToken = accessTokenParser(body);

        // Then modify current request with accessToken
        final midReq = res.middlewareRequest;
        final newReq = midReq.httpRequest;
        final mutation = {
          authenticationHeader: '$authenticationMethod $accessToken',
        };
        Map<String, String>? headers = newReq.headers;
        if (headers != null) {
          headers.addAll(mutation);
        } else {
          headers = mutation;
        }
        midReq.updateHttpRequest(headers: headers);
        final newRes = await httpClient.send(midReq.httpRequest);
        return res.copyWith(httpResponse: res as Response);
      }
    }
    return res;
  }
}
