// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:http/http.dart';
import 'package:wyatt_http_client/src/middleware_client.dart';
import 'package:wyatt_http_client/src/models/middleware_request.dart';
import 'package:wyatt_http_client/src/models/middleware_response.dart';

part 'middleware.dart';
part 'middleware_node.dart';

class Pipeline {
  int _length = 0;
  late final MiddlewareNode begin;
  late final MiddlewareNode end;

  MiddlewareNode get first => begin.child;
  MiddlewareNode get last => end.parent;
  bool get isEmpty => _length == 0;
  bool get isNotEmpty => !isEmpty;
  int get length => _length;

  Pipeline() {
    _initialize();
  }

  Pipeline.fromIterable(Iterable<Middleware?> list) {
    _initialize();
    MiddlewareNode parent = begin;
    for (final element in list) {
      if (element != null) {
        parent = parent.insertAfter(element);
      }
    }
  }

  Pipeline.from(Pipeline pipeline)
      : this.fromIterable(pipeline.middlewares.map((node) => node.middleware));

  void _initialize() {
    _length = 0;
    begin = MiddlewareNode._begin(this);
    end = MiddlewareNode._end(this);
    begin._child = end;
    end._parent = begin;
  }

  Iterable<MiddlewareNode> get middlewares sync* {
    for (var middleware = first;
        middleware != end;
        middleware = middleware.child) {
      yield middleware;
    }
  }

  int indexOf(MiddlewareNode node) {
    int i = -1;
    int j = -1;
    for (final element in middlewares) {
      j++;
      if (element == node) {
        i = j;
        continue;
      }
    }
    return i;
  }

  int indexOfChild(Middleware middleware) {
    int i = -1;
    int j = -1;
    for (final element in middlewares) {
      j++;
      if (element.middleware == middleware) {
        i = j;
        continue;
      }
    }
    return i;
  }

  /// Create new [Pipeline] from first [Middleware] to the specified one.
  Pipeline fromUntil(Middleware middleware, {bool include = false}) {
    final nodes = <Middleware?>[];
    for (final element in middlewares) {
      if (element.middleware != middleware) {
        nodes.add(element.middleware);
      }
      if (element.middleware == middleware) {
        if (include) {
          nodes.add(element.middleware);
        }
        break;
      }
    }
    return Pipeline.fromIterable(nodes);
  }

  Pipeline addMiddleware(Middleware middleware) {
    last.insertAfter(middleware);
    return this;
  }

  void setClient(MiddlewareClient? client) {
    for (var node = first; node != end; node = node.child) {
      node.middleware?.httpClient = client;
    }
  }

  Future<MiddlewareRequest> onRequest(MiddlewareRequest request) async {
    MiddlewareRequest req = request;
    for (var node = first; node != end; node = node.child) {
      req = await node.middleware?.onRequest(req) ?? req;
    }
    return req;
  }

  Future<MiddlewareResponse> onResponse(MiddlewareResponse response) async {
    MiddlewareResponse res = response;
    for (var node = last; node != begin; node = node.parent) {
      res = await node.middleware?.onResponse(res) ?? res;
    }
    return res;
  }

  String getLogic() {
    final req = <String>[];
    final res = <String>[];
    for (final m in middlewares) {
      req.add('${m.middleware}');
      res.insert(0, '${m.middleware}');
    }
    return '[Req] -> ${req.join(' -> ')}\n[Res] -> ${res.join(' -> ')}';
  }

  @override
  String toString() {
    return getLogic();
  }
}
